# ACL lookup

This directory contains files used for the ACL workshop. The goal of
this workshop is to optimize a basic password reset service.

## Description

The password reset service is installed under `/password-reset` and
takes the `password` argument of the query to reset the user
password. The password it looked up in a dictionary file and the
request is denied if the password is found in the dictionary. For
example:

```
/reset-password?password=Secret
```

The original dictionary is a list of regular expressions stored in the
`forbidden-passwords.acl` file.


The password reset is simulated by the `reset-password` backend.

For example sake, we only focus on password, not the user name.

## Performance measurements

The `direct` frontend is here as a reference to see how many requests
per second can be processed. To measure this value, we use `h1load`.

Start the proxy with command:

```
haproxy -f haproxy.cfg
```

and run the injector:

```
taskset -c 1 ./h1load -t 1 -c 64 -d 20 -S 'http://198.18.0.102:8000/reset-password?foo=bar&password=NotInList&bar=foo'
```

Here the reference is 166k req/s. Again this number should not be read
as an absolute value, it will be our reference.

## Original setup

The `original` setup is the actual frontend configuration we want to
optimize. Let's inject some traffic:

```
taskset -c 1 ./h1load -t 1 -c 64 -d 20 -S 'http://198.18.0.102:8001/reset-password?foo=bar&password=NotInList&bar=foo'
```

The results are better than expected: 162k req/s. There is a good
reason for that, HAProxy caches the lookup results in a cache, the
pattern cache. The size of this cache is configurable using the
`tune.pattern.cache-size` directive in the `global` section. To see
the real impact of this cache we should set its size to 0 to disable
it.

The `NO_PATTERN_CACHE` environment variable controls the pattern cache
deactivation. Start HAProxy with following command and run the
injection again:

```
NO_PATTERN_CACHE=1  haproxy -f haproxy.cfg
```

The performances drop to 108k req/s. This shows the impact of the
regular expression lookup.

## First optimization: add path filter

In the `original` frontend, we can see every requests are
analyzed. Frontend `optim-01` adds a filter on the path to only
process `/reset-password`.

Let's do some tests:

```
taskset -c 1 ./h1load -t 1 -c 64 -d 20 -S 'http://198.18.0.102:8002/reset-password?foo=bar&password=NotInList&bar=foo'
```

Results are the same. Let's try a different request:

```
taskset -c 1 ./h1load -t 1 -c 64 -d 20 -S 'http://198.18.0.102:8002/?foo=bar&password=NotInList&bar=foo'
```

Again the filter does not changes anything.

In a filter, order matters, meaning the regular expression is searched
first and for all paths, even those different than `/reset-password`.

## Second optimization: filter path first

Let's invert the path and password checks and try it on `optim-02`.

```
taskset -c 0 ./h1load -t 1 -c 64 -d 20 -S 'http://198.18.0.102:8003/?foo=bar&password=NotInList&bar=foo'
taskset -c 0 ./h1load -t 1 -c 64 -d 20 -S 'http://198.18.0.102:8003/reset-password?foo=bar&password=NotInList&bar=foo'
```

Now `/reset-password` is still slower (104k req/s) but others are way
faster (165k req/s). Since the path check is false, the password is
not evaluated since its result won't change anything in the decision.

Be careful when defining filters, use cheap tests to trim the number
of requests going to CPU consuming ones.

## Third optimization: extract password value

In previous examples the password was searched in the whole query
string (arguments after the question mark in the request). HAProxy has
a neat fetch sample to extract the value from a query argument:
`url_param`. `url_param(password)` returns the value of argument
`password`. The frontend `optim-03` uses that feature.

Instead of looking a complex regular expressions set now we can just
combine them in lines with a `|` (or in regular expression). The
pattern file is `forbidden-passwords-o03.acl`.

```
taskset -c 0 ./h1load -t 1 -c 64 -d 20 -S 'http://198.18.0.102:8004/reset-password?foo=bar&password=NotInList&bar=foo'
```

Results are better: 119k req/s but we can do better.

## Fourth optimization: list of words

Frontend `optim-04` uses a list of password, one per line in file
`forbidden-passwords-o03.acl`. It makes this list easier to maintain.

```
taskset -c 0 ./h1load -t 1 -c 64 -d 20 -S 'http://198.18.0.102:8005/reset-password?foo=bar&password=NotInList&bar=foo'
```

Results drop drastically down to 33k req/s. This is because we are
still using a regular expression lookup over 251 items instead of 26
in `forbidden-passwords-o03.acl`.

A regular expression cannot be indexed. Meaning we have 251 tests to
run for each request. Let's use a string lookup instead.

## Fifth optimization: list of words (string)

Frontend `optim-05` is the same as `optiom-04` but uses string instead
of regular expression to match the password argument.

```
taskset -c 0 ./h1load -t 1 -c 64 -d 20 -S 'http://198.18.0.102:8006/reset-password?foo=bar&password=NotInList&bar=foo'
```

The results are now excellent 158k req/s.

## Bonus 1: case insensitive match

Now let's check password in a case insensitive manner. Frontend
`bonus-01` performs a case insensitive lookup on words.

```
taskset -c 0 ./h1load -t 1 -c 64 -d 20 -S 'http://192.168.255.1:8007/reset-password?foo=bar&password=NotInList&bar=foo'
```

Results are not as good as previous test: 123k req/s. When adding a
case insensitive flag to the lookup (`-i`) the patterns cannot be
indexed. Meaning we are running again 251 comparisons for each
request.

Let's try a better approach.

## Bonus 2: convert password to lower case

the frontend `bonus-02` converts password in lower case and perform an
indexed lookup. It's basically the same as `bonus-01` but instead of
looking into a case insensitive pattern list, the string to match is
converted into lower case and the lookup is done against a pattern
tree.

```
taskset -c 0 ./h1load -t 1 -c 64 -d 20 -S 'http://192.168.255.1:8008/reset-password?foo=bar&password=NotInList&bar=foo'
```

The results are excellent: 153k req/s.

## Bonus 3: use a 10k-word list

Frontend `bonus-3` uses a 9894-word list to check the password.

```
taskset -c 0 ./h1load -t 1 -c 64 -d 20 -S 'http://198.18.0.102:8009/reset-password?foo=bar&password=NotInList&bar=foo'
```

Size doesn't matter, results are stable: 151k req/s

Other tests could be done as a comparison using the 10k-word list:

* `bonus-03.1`: case insensitive (`-i`) lookup: 13k req/s
* `bonus-03.2`: case sensitive regular expressions lookup: 1.3k req/s
* `bonus-03.3`: case insensitive regular expressions lookup: 1.2k req/s
