# Stick tables

This directory contains files used to demonstrate how to configure and
use stick tables within a backend and using the `peers` section.

## Description

Both `haproxy.cfg` and `haproxy-peers.cfg` files provide the same
service. The first one uses backends to define stick tables whereas
the later one uses a `peers` section to achieve the same goal.

Provided service simulates a file download feature with a basic
anti-abuse system with following rules:

* A file can be downloaded only once per hour.
* A client can only:
  * Connect 10 times within a minute.
  * Download 10 different files within a minute.
* If a client generates more than 2 errors (`4xx` or `5xx`) he will be
  banned for 10 minutes.

In this workshop we won't do performance tests but see the difference
between both approaches.

Note: an additional server with IP `198.18.0.103` can be used to test
stick-table replication.

## Backends method

Since we are using the socket we need to specify the `ROOT`
environment variable to force an absolute path. Start HAProxy as
usual:

```
ROOT=$(pwd) NO_PATTERN_CACHE=1 haproxy -f haproxy.cfg
```

Open a second terminal in the root directory to query the stick
tables. Make sure `socat` is installed on your system.

```
socat unix:haproxy.sock - <<< "show table"
# table: per-url, type: binary, size:10240, used:0
# table: per-ip, type: ip, size:10240, used:0
```

Try to request a file:

```
curl http://198.18.0.102:8000/file1
download /file1
```

And check the table content:

```
socat unix:haproxy.sock - <<< "show table per-ip"
# table: per-ip, type: ip, size:10240, used:1
0x55f02ba23040: key=198.18.0.102 use=0 exp=86397369 conn_rate(60000)=1 http_req_rate(60000)=1 http_err_rate(600000)=0 http_fail_rate(600000)=0
```

```
socat unix:haproxy.sock - <<< "show table per-url"
# table: per-url, type: binary, size:10240, used:1
0x55f02baa1540: key=000000002817C73A use=0 exp=86382857 http_req_rate(3600000)=1
```

If you try to download the same file again you will be denied:

```
curlhttp://198.18.0.102:8000/file1
<html><body><h1>403 Forbidden</h1>
Request forbidden by administrative rules.
</body></html>
```


And if you try to download an other file you will be able to download
it:

```
curl http://198.18.0.102:8000/file2
download /file2
```

If you try to download 11 files the same result happens. You can
download the 10 first files and you will be locked out:

```
for i in $(seq 11); do curl http://198.18.0.102:8000/file$i; done
download /file1
download /file2
download /file3
download /file4
download /file5
download /file6
download /file7
download /file8
download /file9
download /file10
curl: (56) Recv failure: Connection reset by peer
```

Now if you have a look at the HAProxy statistics page
(http://localhost:7000) you will see the 2 extra backends (`per-ip`
and `per-url`).


## Peers method

The peers methods offers the exact same serve as the backend method,
except there is no `per-ip` nor `per-url` backends in the statistics
page.

Starts the proxy using this command:

```
ROOT=$(pwd) NO_PATTERN_CACHE=1 haproxy -f haproxy-peers.cfg
```

One other difference is that when you want to see the table content
you have to refer it by prefixing its name with the peers section
name:

```
socat unix:haproxy.sock - <<< "show table my-app/per-ip"
# table: my-app/per-ip, type: ip, size:10240, used:1
0x564ffb9ca840: key=198.18.0.102 use=0 exp=86392233 conn_rate(60000)=11 http_req_rate(60000)=10 http_err_rate(600000)=0 http_fail_rate(600000)=0
```

## Stick table replication

If you have 2 HAProxy instances you can replicate the stick tables
from one node to the other:

On `lb1` (IP: 198.18.0.102):

```
ROOT=$(pwd) NO_PATTERN_CACHE=1 IS_CLUSTER=1 haproxy -f haproxy-peers.cfg -L lb1
```

On `lb2` (IP: 198.18.0.103):

```
ROOT=$(pwd) NO_PATTERN_CACHE=1 IS_CLUSTER=1 haproxy -f haproxy-peers.cfg -L lb2
```

If you are using different IP addresses you can define `PEER_IP_LB1`
and `PEER_IP_LB2` environment variables.


Now query the first server:

```
for i in $(seq 11); do echo http://198.18.0.102:8000/file$i; done | xargs curl
```

And check on both load-balancers the tables content.

On `lb1`:

```
socat unix:haproxy.sock - <<< "show table my-app/per-ip"
# table: my-app/per-ip, type: ip, size:10240, used:1
0x5625146d0ba0: key=198.18.0.102 use=0 exp=86397359 conn_rate(60000)=1 http_req_rate(60000)=11 http_err_rate(600000)=0 http_fail_rate(600000)=0
```

On `lb2`:

```
socat unix:haproxy.sock - <<< "show table my-app/per-ip"
# table: my-app/per-ip, type: ip, size:10240, used:1
0x55c00bf9f040: key=198.18.0.103 use=0 exp=86396767 conn_rate(60000)=1 http_req_rate(60000)=11 http_err_rate(600000)=0 http_fail_rate(600000)=0
```


If you stop HAProxy on `lb` and start it again you can check the table
content has been replicated from `lb2`.
